#include <u.h>
#include <libc.h>
#include <bio.h>
#include "misc.h"

void *
emalloc(ulong n)
{
	void *p = malloc(n);
	if(p==0)
		sysfatal("out of memory");
	return p;
}

void *
ecalloc(ulong n, ulong elsize)
{
	void *p = calloc(n, elsize);
	if(p==0)
		sysfatal("out of memory");
	return p;
}

char *
estrdup(char *str)
{
	char *p;
	p = strdup(str);
	if(p == nil)
		sysfatal("no space for strdup");
	return p;
}

List *
Lpush(List *root, void *data)
{
	List *p = new(List);
	p->data = data;
	p->next = root;
	return p;
}

List*
Lappend(List *p, void *d)
{
	List *q;
	q = (List*) malloc(sizeof(List));
	q->next = p->next;
	q->data = d;
	p->next = q;
	return q;
};

void
Lfree(List *p)
{
	List *q;
	q = p->next;
	free(p);
	while(q != 0){
		p = q->next;
		free(q);
		q = p;
	}
}

Rdbuf *
rdinit(int fd, int size)
{
	Rdbuf *p;
	p = emalloc(sizeof(Rdbuf));
	p->fd = fd;
	p->size = size;
	p->bp = emalloc(size);
	p->sp = p->cp = p->ep = nil;
	return p;
};

void
rdfree(Rdbuf *rb) /* left for compatibility */
{
	free(rb->bp);
	free(rb);
}

int
rdlnlen(Rdbuf *rb)
{
	char *cp = rb->cp;
	if(*(cp - 1) == 0)
		return cp - rb->sp -1;
	return cp - rb->sp;
}

/*	Bio function Brdstr() is slow
*	Brdline() is double times better than Brdstr(), but has a problem with EOF.
*	rdline() below is similar to Brdline() but works reasonably at EOF. */
char *
rdline(Rdbuf *rb)
{
	int fd, size;
	long n;
	ulong m;
	char *bp, *sp, *cp, *ep, *ep0;
	fd = rb->fd;
	size = rb->size;
	bp = rb->bp;
	sp = rb->sp;
	cp = rb->cp;
	ep = rb->ep;
	ep0 = bp + size;

	if(sp == nil){
		n = read(fd, bp, size);
		if(n < 0)
			sysfatal("read error %r");
		if(n == 0) /* eof */
			return nil;
		cp = bp;
		ep = cp + n;
	}

	sp = cp;
L1:
	while(cp < ep && *cp != '\n')
		cp++;
	if(cp == ep){
		m = cp - sp;
		if(ep == ep0){
			m = cp - sp;
			if(m == size){
				size *= 2;
				bp = realloc(bp, size);
				if(bp == nil)
					sysfatal("error: realloc");
				sp = bp + (sp - rb->bp);
				cp = bp + (cp - rb->bp);
				ep = bp + (ep - rb->bp);
				ep0 = bp + size;
				rb->size = size;
				rb->bp = bp;
				rb->sp = sp;
				rb->cp = cp;
				rb->ep = ep;
				/* then m != size, we can continue */
			}
			memcpy(bp, sp, m);
			sp = bp;
			cp = bp + m;
		}
		n = read(fd, cp, ep0 - cp);
		if(n == 0) {/* eof */
			if(m){
				ep = cp;
				*cp = 0;
				rb->sp = sp;
				rb->cp = cp;
				rb->ep = ep;
				return sp;
			}
			return nil;
		}
		if(n < 0)
			sysfatal("read error %r");
		ep = cp + n;
		goto L1;
	}
	/*	here, *p is NL */
	*cp++ = 0;
	rb->sp = sp;
	rb->cp = cp;
	rb->ep = ep;
	return sp;
}


/*	Why getfields() is slower than tokenize() ?
*	this is because getfields() allows utf coded chars
*	in the separater. it would be much simpler and
*	faster one if the separater is a single char and ascii
*
*	sep is 0, then split returns 1. args[0] is the str	*/

int
split(char *str, char **args, int max, int sep)
{
	char *p,*s;
	int n;
	n = 0;
	for(s = p = str; *p; p++){
		if(*p == sep){
			*p = 0;
			args[n++] = s;
			if(n == max)
				return n;
			s = p + 1;
		}
	}
	args[n++] = s; /* the rest */
	return n;
}

#ifdef XXX
void
strtolower(char *s)
{
	while(*s){
		*s = tolower(*s);
		s++;
	}
}
#endif

void
strtolower(char *s)
{
	Rune r0, r1;
	int n0,n1;
	char *s0 = s;
	char *t = s;

	for(; *s; s += n0, t+= n1){
		n0 = chartorune(&r0,s);
		if(r0 == 0xfffd){ /* rune error, skip this rune */
			n1 = 0;
			continue;
		}
		r1 = tolowerrune(r0);
		n1 = runetochar(t,&r1);
		/*	some runes are converted to smaller size.
		*	such examples are: 0x0130(İ) to 0x0069(i)	*/
		/*	we don't support inflation case.
		*	looking /sys/src/libc/port/runetype.c, we have no such cases. */
		if(n0 < n1)
			sysfatal("strtolower: something wrong %04x  %04x %d %d %s",r0,r1,n0,n1,s0);
	}
	*t = 0;
}

char *
skip(char *s, char *p)
{
	while(*s){
		if(!strchr(p,*s))
			break;
		s++;
	}
	return s;
}

char *
skipto(char *s, char *p)
{
	while(*s){
		if(strchr(p,*s))
			break;
		s++;
	}
	return s;
}

int
chkutf(char *s)
/*	check the validity of utf string
 *	return
 *	0: valid
 *	1: invalid
 */
{
	int n;
	Rune r;
	for(;*s;s += n) {
		n = chartorune(&r, s);
		if(r == Runeerror)
			return 1;
	}
	return 0;
}


char *
skipnrune(char *p, int n)
/*	utfnrune() returns number of byte to skip n utf chars
*	p is a pointer to utf string
*	valid utf-8 is assumed
*
*	is it better to apply utfnlen() in libc? */
{
	char *e;
	e = p + strlen(p);
	while(n > 0 && p < e){
		if((*p & 0x80) == 0) // 0bbbbbbb
			p += 1;
		else
		if((*p & 0xe0) == 0xc0) // 110bbbbb, 10bbbbbb
			p += 2;
		else
		if((*p & 0xf0) == 0xe0) // 1110bbbb, 10bbbbbb, 10bbbbbb,
			p += 3;
		else
		if((*p & 0xf0) == 0xf0) // 1111bbbb, 10bbbbbb, 10bbbbbb, 10bbbbbb,
			p += 4;
		else // need for safety
			p++;
		/* NB: last case is actually appeared in my documents */
		n--;
	}
	/*	if p>e then we should return e */
	return p>e?e:p;
}

